#!/usr/bin/env python
# =============================================================================
# Copyright [2013] [Kevin Carter]
# License Information :
# This software has no warranty, it is provided 'as is'. It is your
# responsibility to validate the behavior of the routines and its accuracy
# using the code provided. Consult the GNU General Public license for further
# details (see GNU General Public License).
# http://www.gnu.org/licenses/gpl.html
# =============================================================================
import json
import sys

import turbolift as turbo
from turbolift import arguments
from turbolift import load_constants
from turbolift.logger import logger


def run_turbolift():
    """This is the run section of the application Turbolift."""

    if len(sys.argv) <= 1:
        arguments.get_help()
        raise SystemExit('Give me something to do and I will do it')
    else:
        args = arguments.get_args()
        log = logger.LogSetup(
            debug_logging=args.get('debug', False),
            log_dir=args.get('log_location', '/var/log'),
            log_name=args.get('log_file')
        ).default_logger(enable_stream=args.get('log_streaming'))

        log.debug('set arguments [ %s ]', json.dumps(args, indent=2))

        import turbolift.utils.basic_utils as basic
        args = basic.dict_pop_none(dictionary=args)
        load_constants(args=args)

        try:
            from turbolift import worker
            worker.start_work()
        except KeyboardInterrupt:
            turbo.emergency_kill(reclaim=True)
        finally:
            if args.get('quiet') is not True:
                print('All Done!')
            log.info('Job Finished.')


if __name__ == "__main__":
    run_turbolift()



# =============================================================================
# Copyright [2013] [Kevin Carter]
# License Information :
# This software has no warranty, it is provided 'as is'. It is your
# responsibility to validate the behavior of the routines and its accuracy
# using the code provided. Consult the GNU General Public license for further
# details (see GNU General Public License).
# http://www.gnu.org/licenses/gpl.html
# =============================================================================
import logging
from logging import handlers
import os


"""Example Usage:
>>> from cloudlib import logger
>>> log = logger.LogSetup()
>>> log.default_logger(name='test_logger')

>>> # The following can then be placed in any module that you like
>>> # Just use the name of the logger you created
>>> from cloudlib import logger
>>> LOG = logger.getLogger(name='test_logger')
>>> LOG.info('This is a test message')
"""


def getLogger(name):
    """Return a logger from a given name.

    If the name does not have a log handler, this will create one for it based
    on the module name which will log everything to a log file in a location
    the executing user will have access to.

    :param name: ``str``
    :return: ``object``
    """
    log = logging.getLogger(name=name)
    for handler in log.handlers:
        if name == handler.name:
            return log
    else:
        return LogSetup().default_logger(name.split('.')[0])


class LogSetup(object):

    def __init__(self, max_size=500, max_backup=5, debug_logging=False,
                 log_dir='/var/log', log_name=__name__):
        """Setup Logging.

        :param max_size: ``int``
        :param max_backup: ``int``
        :param debug_logging: ``bol``
        :param log_dir: ``str``
        :param log_name: ``str``
        """
        self.max_size = (max_size * 1024 * 1024)
        self.max_backup = max_backup
        self.debug_logging = debug_logging
        self.format = None
        self.name = log_name
        self.log_dir = log_dir

    def default_logger(self, enable_stream=False, enable_file=True):
        """Default Logger.

        This is set to use a rotating File handler and a stream handler.
        If you use this logger all logged output that is INFO and above will
        be logged, unless debug_logging is set then everything is logged.
        The logger will send the same data to a stdout as it does to the
        specified log file.

        You can disable the default handlers by setting either `enable_file` or
        `enable_stream` to `False`

        :param enable_stream: ``bol``
        :param enable_file: ``bol``
        :return: ``object``
        """
        if self.format is None:
            self.format = logging.Formatter(
                '%(asctime)s - %(module)s:%(levelname)s => %(message)s'
            )

        log = logging.getLogger('turbolift')

        if enable_file is True:
            log_file_name = self.name
            if not log_file_name.endswith('log'):
                log_file_name = '%s.log' % log_file_name

            log_file = self.return_logfile(
                filename=log_file_name, log_dir=self.log_dir
            )
            file_handler = handlers.RotatingFileHandler(
                filename=log_file,
                maxBytes=self.max_size,
                backupCount=self.max_backup
            )
            self.set_handler(log, handler=file_handler)

        if enable_stream is True or self.debug_logging is True:
            stream_handler = logging.StreamHandler()
            self.set_handler(log, handler=stream_handler)

        log.debug('Logger [ %s ] loaded', self.name)
        return getLogger('turbolift')

    def set_handler(self, log, handler):
        """Set the logging level as well as the handlers.

        :param log: ``object``
        :param handler: ``object``
        """

        if self.debug_logging is True:
            log.setLevel(logging.DEBUG)
            handler.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)
            handler.setLevel(logging.INFO)

        handler.name = 'turbolift'
        handler.setFormatter(self.format)
        log.addHandler(handler)

    @staticmethod
    def return_logfile(filename, log_dir):
        """Return a path for logging file.

        If ``log_dir`` exists and the userID is 0 the log file will be written
        to the provided log directory. If the UserID is not 0 or log_dir does
        not exist the log file will be written to the users home folder.

        :param filename: ``str``
        :param log_dir: ``str``
        :return: ``str``
        """
        user = os.getuid()
        home = os.getenv('HOME')
        default_log_location = os.path.join(home, filename)
        if not user == 0:
            return default_log_location
        else:
            if os.path.isdir(log_dir):
                return os.path.join(log_dir, filename)
            else:
                return default_log_location


# =============================================================================
# Copyright [2013] [Kevin Carter]
# License Information :
# This software has no warranty, it is provided 'as is'. It is your
# responsibility to validate the behavior of the routines and its accuracy
# using the code provided. Consult the GNU General Public license for further
# details (see GNU General Public License).
# http://www.gnu.org/licenses/gpl.html
# =============================================================================
import json
import traceback
import urlparse

import turbolift as turbo
import turbolift.utils.http_utils as http

from turbolift import ARGS
from turbolift import info
from turbolift.logger import logger


LOG = logger.getLogger('turbolift')
AUTH_VERSION_MAP = {
    'V1': 'v1.0',
    'v1': 'v1.0',
    'v1.0': 'v1.0',
    'V1.0': 'v1.0',
    '1': 'v1.0',
    '1.0': 'v1.0',
    'V2': 'v2.0',
    'v2': 'v2.0',
    'v2.0': 'v2.0',
    'V2.0': 'v2.0',
    '2': 'v2.0',
    '2.0': 'v2.0',
}


def auth_request(url, headers=None, body=None):
    """Perform auth request for token."""
    headers = headers or {}
    body = body or json.dumps({})
    if get_authversion() == 'v1.0':
        # shouldnt need to allow 'stream'
        return http.get_request(url, headers, None)
    else:
        # shouldnt need to allow 'rpath'
        return http.post_request(url, headers, body=body)


def get_headers():
    """Setup headers for authentication request."""

    if get_authversion() == 'v1.0':
        if all([ARGS.get(s) for s in ('st_user', 'st_key')]):
            return {
                'X-Auth-User': ARGS['st_user'],
                'X-Auth-Key': ARGS['st_key']
            }
        else:
            LOG.error(traceback.format_exc())
            raise AttributeError('Missing Password, APIKey, Token, or User')


def parse_reqtype():
    """Setup our Authentication POST.

    username and setup are only used in APIKEY/PASSWORD Authentication
    """

    if get_authversion() == 'v1.0':
        return
    else:
        setup = {'username': ARGS.get('os_user')}
        if ARGS.get('os_token') is not None:
            auth_body = {'auth': {'token': {'id': ARGS.get('os_token')}}}
        elif ARGS.get('os_password') is not None:
            prefix = 'passwordCredentials'
            setup['password'] = ARGS.get('os_password')
            auth_body = {'auth': {prefix: setup}}
        elif ARGS.get('os_apikey') is not None:
            prefix = 'RAX-KSKEY:apiKeyCredentials'
            setup['apiKey'] = ARGS.get('os_apikey')
            auth_body = {'auth': {prefix: setup}}
        else:
            LOG.error(traceback.format_exc())
            raise AttributeError('No Password, APIKey, or Token Specified')

        if ARGS.get('os_tenant'):
            auth_body['auth']['tenantName'] = ARGS.get('os_tenant')

        LOG.debug('AUTH Request Type > %s', auth_body)
        return auth_body


def get_authversion():
    """Get or infer the auth version."""

    authversion = ARGS.get('auth_version')
    authversion = AUTH_VERSION_MAP.get(authversion) or authversion
    if authversion:
        supported = ['v1.0', 'v2.0']
        if authversion not in supported:
            raise ValueError("Auth Version must be one of %s."
                             % supported)
    else:
        # infer version if possible else v2.0
        if any((ARGS.get(s) for s in ('st_auth', 'st_user', 'st_key'))):
            authversion = 'v1.0'
        elif '/v1.0' in ARGS.get('os_auth_url', ''):
            raise ValueError("Specify v1 auth endpoint with 'st_auth'"
                             "instead of 'os_auth_url'")
        elif '/v2.0' in ARGS.get('st_auth', ''):
            raise ValueError("Specify v2 auth endpoint with 'os_auth_url'"
                             "instead of 'st_auth'")
        else:
            authversion = 'v2.0'

    if authversion == 'v1.0':
        if not ARGS.get('st_auth'):
            raise AttributeError("Specify the v1 auth endpoint "
                                 "with 'st_auth'")
    ARGS['auth_version'] = authversion
    return authversion


def get_surl(region, cf_list, lookup):
    """Lookup a service URL.

    :param region:
    :param cf_list:
    :param lookup:
    :return net:
    """

    for srv in cf_list:
        region_get = srv.get('region')
        lookup_get = srv.get(lookup)
        if any([region in region_get, region.lower() in region_get]):
            if lookup_get is None:
                return None
            else:
                return http.parse_url(url=lookup_get)
    # TODO(samstav): No break statement in for loop...
    # remove 'else'
    else:
        raise turbo.SystemProblem(
            'Region "%s" was not found in your Service Catalog.' % region
        )


def _parse_v2_auth_response(auth_response):
    """Parse the auth response and return the tenant, token, and username.

    :param auth_response: the full object returned from an auth call
    :returns: tuple (token, tenant, username, internalurl, externalurl, cdnurl)
    """
    try:
        auth_response = auth_response.json()
    except AttributeError:
        pass
    else:
        LOG.debug('Authentication Response Body %s', auth_response)

    def _service_ep(scat, types_list):
        for srv in scat:
            if srv.get('name') in types_list:
                index_id = types_list.index(srv.get('name'))
                index = types_list[index_id]
                if srv.get('name') == index:
                    return srv.get('endpoints')
        else:
            return None

    access = auth_response.get('access')
    token = access.get('token').get('id')

    if 'tenant' in access.get('token'):
        tenant = access.get('token').get('tenant').get('name')
        user = access.get('user').get('name')
    elif 'user' in access:
        tenant = None
        user = access.get('user').get('name')
    else:
        LOG.error('No Token Found to Parse.\nHere is the DATA: %s\n%s',
                  auth_response, traceback.format_exc())
        raise turbo.NoTenantIdFound('When attempting to grab the '
                                    'tenant or user nothing was found.')

    if ARGS.get('os_rax_auth') is not None:
        region = ARGS.get('os_rax_auth')
    elif ARGS.get('os_hp_auth') is not None:
        if ARGS.get('os_tenant') is None:
            raise turbo.NoTenantIdFound(
                'You need to have a tenant set to use HP Cloud'
            )
        region = ARGS.get('os_hp_auth')
    elif ARGS.get('os_region') is not None:
        region = ARGS.get('os_region')
    else:
        raise turbo.SystemProblem('No Region Set')

    scat = access.pop('serviceCatalog')

    cfl = _service_ep(scat, info.__srv_types__)
    cdn = _service_ep(scat, info.__cdn_types__)

    if cfl is not None:
        inet = get_surl(region=region, cf_list=cfl, lookup='internalURL')
        enet = get_surl(region=region, cf_list=cfl, lookup='publicURL')
    else:
        need_tenant = ' Maybe you need to specify "os-tenant"?'
        gen_message = ('No Service Endpoints were found for use '
                       'with Swift. If you have Swift available to you, '
                       'Check Your Credentials and/or Swift\'s '
                       'availability using Token Auth.')
        if ARGS.get('os_tenant') is None:
            gen_message += need_tenant
        raise turbo.SystemProblem(gen_message)

    if cdn is not None:
        cnet = get_surl(region=region, cf_list=cdn, lookup='publicURL')
    else:
        cnet = None

    return token, tenant, user, inet, enet, cnet, cfl


def parse_auth_response(auth_response):
    """Parse the auth response and return the tenant, token, and username.

    :param auth_response: the full object returned from an auth call
    :returns: tuple (token, tenant, username, internalurl, externalurl, cdnurl)
    """

    if get_authversion() == 'v1.0':
        LOG.debug('Authentication Response Headers %s', auth_response.headers)
        token = auth_response.headers['x-auth-token']

        inet = urlparse.urlparse(auth_response.headers['x-storage-url'])
        acfep = inet
        enet = inet
        cnet = None
        tenant = None
        user = ARGS['st_user']
        return token, tenant, user, inet, enet, cnet, acfep
    else:
        return _parse_v2_auth_response(auth_response)


def _parse_v2_region():
    """Pull region/auth url information from context."""
    if ARGS.get('os_rax_auth'):
        region = ARGS.get('os_rax_auth')
        auth_url = 'identity.api.rackspacecloud.com/v2.0/tokens'
        if region is 'LON':
            return ARGS.get('os_auth_url', 'https://lon.%s' % auth_url)
        elif region.lower() in info.__rax_regions__:
            return ARGS.get('os_auth_url', 'https://%s' % auth_url)
        else:
            raise turbo.SystemProblem('No Known RAX Region Was Specified')
    elif ARGS.get('os_hp_auth'):
        region = ARGS.get('os_hp_auth')
        auth_url = 'https://%s.identity.hpcloudsvc.com:35357/v2.0/tokens'
        if region.lower() in info.__hpc_regions__:
            return ARGS.get('os_auth_url', auth_url % region)
        else:
            raise turbo.SystemProblem('No Known HP Region Was Specified')
    elif 'os_auth_url' in ARGS:
        auth_url = ARGS.get('os_auth_url')
        if not auth_url.endswith('/tokens'):
            auth_url = '%s/tokens' % auth_url
        return auth_url
    else:
        raise turbo.SystemProblem(
            'You Are required to specify an Auth URL, Region or Plugin'
        )


def parse_region():
    """Pull region/auth url information from context."""

    authversion = get_authversion()

    if authversion == 'v1.0':
        return ARGS['st_auth']
    else:
        return _parse_v2_region()
